/* $NetBSD$ */

/*
 * File "hashtable.c" is part of the UDFclient toolkit and part of Artdraw.
 * File $Id: hashtable.c,v 1.9 2003/09/14 14:18:59 imago Exp $ $Name: release-ro1 $
 *
 * Copyright (c) 2003 Reinoud Zandijk <reinoud@netbsd.org>
 * All rights reserved.
 *
 * The UDFclient toolkit is distributed under the Clarified Artistic Licence.
 * A copy of the licence is included in the distribution as
 * `LICENCE.clearified.artistic' and a copy of the licence can also be
 * requested at the GNU foundantion's website.
 *
 * Visit the UDFclient toolkit homepage http://www.13thmonkey.org/udfclient/
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <strings.h>
#include "hashtable.h"


static struct hash_entry *hash_entry_pool = NULL;


static struct hash_entry *alloc_hash_entry(void ) {
	struct hash_entry *entry;

	entry = hash_entry_pool;
	if (entry) {
		hash_entry_pool = entry->next;
		entry->next = NULL;
	} else {
		entry = calloc(1, sizeof(struct hash_entry));
	};
	return entry;
}


static void recycle_hash_entry(struct hash_entry *entry) {
	entry->next = hash_entry_pool;
	entry->seq_next = NULL;
	entry->seq_prev = NULL;
	hash_entry_pool = entry;
}


static int optimise_number_of_buckets(int num_buckets) {
	int power;

	num_buckets <<= 1;
	for(power=0; num_buckets; power++) {
		num_buckets >>= 1;
	};
	power--;
	return (1 << power);
}


int hash_init_generic_table(struct hash_table *table, int num_buckets, hash_compare compare, hash_calc_hash calc) {
	if (!table)	return EFAULT;
	if (!compare)	return EFAULT;
	if (!calc)	return EFAULT;

#if 0
	if (table->num_entries || table->seq_head || table->entries) {
		fprintf(stderr, "hashtable : warning reinitialising hashtable; deleting all entries\n");
		hash_delete_all(table);
		free(table->entries);
	};
#endif

	/* try to find a reasonable number of buckets */
	num_buckets = optimise_number_of_buckets(num_buckets);

	table->entries = calloc(num_buckets, sizeof(struct hash_entry *));
	if (!table->entries) return ENOMEM;

	table->bucket_mask = num_buckets-1;
	table->compare = compare;
	table->calc = calc;

	table->num_entries = 0;
	table->seq_head = NULL;

	return 0;
}


int hash_add_entry(struct hash_table *table, void *key, void *entry) {
	uint32_t hash, bucket;
	struct hash_entry *hash_entry;

	hash = table->calc(key);
	bucket = hash & table->bucket_mask;
	assert(bucket <= table->bucket_mask);

	/* allocate hash_entry */
	hash_entry = alloc_hash_entry();
	if (!hash_entry) return ENOMEM;

	/* fill in hash_entry */
	hash_entry->hash_value = hash;
	hash_entry->key = key;
	hash_entry->entry = entry;

	hash_entry->next = table->entries[bucket];
	table->entries[bucket] = hash_entry;

	/* add to the linked list */
	hash_entry->seq_prev  = NULL;
	hash_entry->seq_next  = table->seq_head;
	if (table->seq_head) table->seq_head->seq_prev = hash_entry;
	table->seq_head = hash_entry;

	table->num_entries++;

	return 0;
}


int hash_delete_entry(struct hash_table *table, void *key) {
	uint32_t hash, bucket;
	struct hash_entry *hash_entry, *former_entry;

	hash = table->calc(key);
	bucket = hash & table->bucket_mask;
	assert(bucket <= table->bucket_mask);

	/* search in table bucket chain */
	hash_entry = table->entries[bucket];
	former_entry = NULL;
	while (hash_entry) {
		if (hash_entry->hash_value == hash) {
			if (table->compare(hash_entry->key, key) ==0) {
				if (former_entry) {
					/* delete entry from bucket chain list */
					former_entry->next = hash_entry->next;
				} else {
					/* delete from front */
					table->entries[bucket] = hash_entry->next;
				};

				/* move on `head' if it is the seq_head */
				if (hash_entry == table->seq_head) {
					table->seq_head = hash_entry->seq_next;
				};
				/* delete entry from sequential chain */
				if (hash_entry->seq_prev) hash_entry->seq_prev->seq_next = hash_entry->seq_next;
				if (hash_entry->seq_next) hash_entry->seq_next->seq_prev = hash_entry->seq_prev;
				if (table->seq_head == hash_entry) table->seq_head = hash_entry->seq_next;

				recycle_hash_entry(hash_entry);
				table->num_entries--;
				return 0;
			};
		};
		former_entry = hash_entry;
		hash_entry = hash_entry->next;
	};
	/* not found ?? */
	return ENOENT;
}


int hash_delete_entry_by_record(struct hash_table *table, struct hash_entry *entry) {
	/* this could be done more gently */
	return hash_delete_entry(table, entry->key);
}


int hash_lookup_entry(struct hash_table *table, void *key, void **entry) {
	uint32_t hash, bucket;
	struct hash_entry *hash_entry;

	hash = table->calc(key);
	bucket = hash & table->bucket_mask;
	assert(bucket <= table->bucket_mask);

	/* search in chain */
	hash_entry = table->entries[bucket];
	while (hash_entry) {
		if (hash_entry->hash_value == hash) {
			if (table->compare(hash_entry->key, key) ==0) {
				*entry = hash_entry->entry;
				return 0;
			};
		};
		hash_entry = hash_entry->next;
	};
	/* not found ?? */
	*entry = NULL;
	return ENOENT;
}


static int hash_string_compare(void *v1, void *v2) {
	char *cv1, *cv2;

	cv1 = (char *) v1;
	cv2 = (char *) v2;

	return strcmp(cv1, cv2);
}


static uint32_t hash_string_hash(void *v1) {
	uint32_t hash;
	char *cv1;

	cv1 = (char *) v1;
	hash = 0;
	while (*cv1) {
		hash = (hash * 7) + (*cv1);
		cv1++;
	};
	return hash;
}


static int hash_numeric_compare(void *v1, void *v2) {
	uint32_t *nv1, *nv2;

	nv1 = (uint32_t *) v1;
	nv2 = (uint32_t *) v2;

	return (*nv1 - *nv2);		/* 0 flags success */
}


static uint32_t hash_numeric_hash(void *v1) {
	return *((uint32_t *) v1);
}


int hash_init_name_table(struct hash_table *table, int num_buckets) {
	return hash_init_generic_table(table, num_buckets, hash_string_compare, hash_string_hash);
}


int hash_init_numeric_table(struct hash_table *table, int num_buckets) {
	return hash_init_generic_table(table, num_buckets, hash_numeric_compare, hash_numeric_hash);
}


int hash_delete_all(struct hash_table *table) {
	struct hash_entry *hash_entry, *next_entry;

	if (table->num_entries) {
		hash_entry = table->seq_head;
		while (hash_entry) {
			next_entry = hash_entry->seq_next;
			recycle_hash_entry(hash_entry);
			hash_entry = next_entry;
		};
		table->seq_head = NULL;
		table->num_entries = 0;
		bzero(table->entries, (table->bucket_mask+1)*sizeof(struct hash_entry *));
	};
	return 0;
}


/* end of hashtable.c */

