/*	$NetBSD: scsi_subr.c,v 1.7 2002/10/08 20:17:06 soren Exp $	*/

/*-
 * Copyright (c) 1998 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Charles M. Hannum; Jason R. Thorpe of the Numerical Aerospace
 * Simulation Facility, NASA Ames Research Center.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *        This product includes software developed by the NetBSD
 *        Foundation, Inc. and its contributors.
 * 4. Neither the name of The NetBSD Foundation nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Small changes and Linux implementation by Reinoud Zandijk <reinoud@netbsd.org>
 *
 */

/*
 * SCSI support subroutines.
 *
 * XXX THESE SHOULD BE IN A LIBRARY!
 */

#include <sys/param.h>
#include <sys/ioctl.h>
#include <err.h> 
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/types.h>
#ifndef NO_STDINT
#       include <stdint.h>
#endif

#include "scsi_sense.h"
#include "scsilib.h"

#define	STRVIS_ISWHITE(x) ((x) == ' ' || (x) == '\0' || (x) == (u_char)'\377')

int scsilib_verbose = 0;


#ifdef UDF_SCSIPI
#include <sys/scsiio.h>
#include <dev/scsipi/scsipi_all.h>

int
scsi_command(fd, cmd, cmdlen, data, datalen, timeout, flags)
	int fd;
	void *cmd;
	size_t cmdlen;
	void *data;
	size_t datalen;
	int timeout, flags;
{
	scsireq_t req;

	memset(&req, 0, sizeof(req));

	memcpy(req.cmd, cmd, cmdlen);
	req.cmdlen = cmdlen;
	req.databuf = data;
	req.datalen = datalen;
	req.timeout = timeout;
	req.flags = flags;
	req.senselen = SENSEBUFLEN;

	if (ioctl(fd, SCIOCCOMMAND, &req) == -1)
		err(1, "SCIOCCOMMAND");

	if (req.retsts == SCCMD_OK)
		return 0;

	/* Some problem; report it and exit. */
	if (req.retsts == SCCMD_TIMEOUT) {
		if (scsilib_verbose)
			fprintf(stderr, "%s: SCSI command timed out\n", dvname);
		return EAGAIN;
	} else if (req.retsts == SCCMD_BUSY) {
		if (scsilib_verbose)
			fprintf(stderr, "%s: device is busy\n", dvname);
		return EBUSY;
	} else if (req.retsts == SCCMD_SENSE) {
		if (scsilib_verbose)
			scsi_print_sense((char *) dvname, req.cmd, req.cmdlen, req.sense, req.senselen_used, 1);
		return EIO;
	} else
		if (scsilib_verbose)
			fprintf(stderr, "%s: device had unknown status %x\n", dvname,
		  	  req.retsts);

	return EFAULT;
}


int
scsi_mode_sense(fd, pgcode, pctl, buf, len)
	int fd;
	u_int8_t pgcode, pctl;
	void *buf;
	size_t len;
{
	struct scsipi_mode_sense cmd;

	memset(&cmd, 0, sizeof(cmd));
	memset(buf, 0, len);

	cmd.opcode = MODE_SENSE;
	cmd.page = pgcode | pctl;
	cmd.u_len.scsi.length = len;

	return scsi_command(fd, &cmd, sizeof(cmd), buf, len, 10000, SCCMD_READ);
}


int
scsi_mode_select(fd, byte2, buf, len)
	int fd;
	u_int8_t byte2;
	void *buf;
	size_t len;
{
	struct scsipi_mode_select cmd;

	memset(&cmd, 0, sizeof(cmd));

	cmd.opcode = MODE_SELECT;
	cmd.byte2 = SMS_PF | byte2;
	cmd.u_len.scsi.length = len;

	return scsi_command(fd, &cmd, sizeof(cmd), buf, len, 10000, SCCMD_WRITE);
}


int
scsi_request_sense(fd, buf, len)
	int fd;
	void *buf;
	size_t len;
{
	struct scsipi_sense cmd;

	memset(&cmd, 0, sizeof(cmd));
	memset(buf, 0, len);

	cmd.opcode = REQUEST_SENSE;
	cmd.length = len;
 	return scsi_command(fd, &cmd, sizeof(cmd), buf, len, 10000, SCCMD_READ);
}


int scsi_check_for_scsi(int fhandle) {
	struct scsi_addr	saddr;
	return ioctl(fhandle, SCIOCIDENTIFY, &saddr);
}


int scsi_identify(int fhandle, struct scsi_addr *saddr) {
	return ioctl(fhandle, SCIOCIDENTIFY, saddr);
}

#endif


#ifdef UDF_LINUX_SCSI

#include <scsi/sg.h>

#define SENSEBUFLEN 48

int
scsi_command(fd, cmd, cmdlen, data, datalen, timeout, flags)
	int fd;
	void *cmd;
	size_t cmdlen;
	void *data;
	size_t datalen;
	int timeout, flags;
{
	struct sg_io_hdr req;
	uint8_t sense_buffer[SENSEBUFLEN];

	bzero(&req, sizeof(req));
	if (flags == SG_DXFER_FROM_DEV) bzero(data, datalen);

	req.interface_id    = 'S';
	req.dxfer_direction = flags;
	req.cmd_len         = cmdlen;
	req.mx_sb_len       = SENSEBUFLEN;
	req.iovec_count     = 0;
	req.dxfer_len       = datalen;
	req.dxferp          = data;
	req.cmdp            = cmd;
	req.sbp             = sense_buffer;
	req.flags           = 0;
	req.timeout         = timeout;

	errno = ioctl(fd, SG_IO, &req);

	if (req.status) {
		/* Is this OK? */
		if (scsilib_verbose) {
			scsi_print_sense((char *) dvname, cmd, cmdlen, sense_buffer, req.sb_len_wr, 1);
		};
	};

	return errno;
}


int
scsi_mode_sense(fd, pgcode, pctl, buf, len)
	int fd;
	u_int8_t pgcode, pctl;
	void *buf;
	size_t len;
{
	return ENODEV;
}


int
scsi_mode_select(fd, byte2, buf, len)
	int fd;
	u_int8_t byte2;
	void *buf;
	size_t len;
{
	return ENODEV;
}


int
scsi_request_sense(fd, buf, len)
	int fd;
	void *buf;
	size_t len;
{
	return ENODEV;
}


int scsi_check_for_scsi(int fhandle) {
	struct scsi_addr saddr;

	return scsi_identify(fhandle, &saddr);
}


int scsi_identify(int fhandle, struct scsi_addr *saddr) {
	struct sg_scsi_id sg_scsi_id;

	errno = ioctl(fhandle, SG_GET_SCSI_ID, &sg_scsi_id);
	
	saddr->type = TYPE_SCSI;	/* how to recognize ATAPI? -> SG_EMULATED_HOST */
	saddr->addr.scsi.scbus  = sg_scsi_id.channel; //host_no;	/* or channel? */
	saddr->addr.scsi.target = sg_scsi_id.scsi_id;
	saddr->addr.scsi.lun    = sg_scsi_id.lun;

	return errno;
}

#endif

