/* $NetBSD$ */

/*
 * File "udf_verbose.c" is part of the UDFclient toolkit.
 * File $Id: udf_verbose.c,v 1.40 2003/09/14 14:19:02 imago Exp $ $Name: release-ro1 $
 *
 * Copyright (c) 2003 Reinoud Zandijk <reinoud@netbsd.org>
 * All rights reserved.
 *
 * The UDFclient toolkit is distributed under the Clarified Artistic Licence.
 * A copy of the licence is included in the distribution as
 * `LICENCE.clearified.artistic' and a copy of the licence can also be
 * requested at the GNU foundantion's website.
 *
 * Visit the UDFclient toolkit homepage http://www.13thmonkey.org/udfclient/
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include <sys/types.h>
#ifndef NO_STDINT
#	include <stdint.h>
#endif

#include "hashtable.h"
#include "udf.h"
#include "udf_bswap.h"


extern int errno;


void udf_dump_id(char *prefix, int len, char *id, struct charspec *chsp);
void udf_dump_long_ad(char *prefix, struct long_ad *adr);
void udf_dump_descriptor(union dscrptr *dscrpt);
void udf_dump_vat_table(struct udf_part_mapping *udf_part_mapping);


void udf_dump_unimpl(union dscrptr *dscrpt) {
	dscrpt = dscrpt;
	fprintf(stderr, "\t\t(unimplemented dump)\n");
}


void udf_dump_desc(struct desc_tag *tag) {
	printf("\tTAG: descriptor %d, serial_num %d at sector %d\n",
			udf_rw16(tag->id), udf_rw16(tag->serial_num), udf_rw32(tag->tag_loc));
}


void udf_dump_anchor(struct anchor_vdp *vdp) {
	printf("\t\tAnchor\n");
	printf("\t\t\tMain    volume descriptor set at %d for %d bytes\n",
			udf_rw32(vdp->main_vds_ex.loc),    udf_rw32(vdp->main_vds_ex.len));
	printf("\t\t\tReserve volume descriptor set at %d for %d bytes\n",
			udf_rw32(vdp->reserve_vds_ex.loc), udf_rw32(vdp->reserve_vds_ex.len));
}


void udf_dump_disc_anchors(struct udf_discinfo *disc) {
	int session;

	printf("\nUDF Dump of disc in device %s\n", disc->dev_name);
	printf("UDF sessions : ");
	for (session = 0; session < disc->num_sessions; session++) {
		if (disc->session_is_UDF[session]) {
			printf("Yes");
			if (disc->session_quirks[session] & CD_SESS_QUIRK_SESSION_LOCAL) {
				printf("(local)");
			};
			printf(" ");
		} else {
			printf("No ");
		};
	};
	printf("\n\n");
	UDF_VERBOSE_TABLES(
		struct udf_session *udf_session;
		/* Dump anchors */
		udf_session = disc->sessions;
		while (udf_session) {
			printf("UDF session %d (lba %d + %d sectors) anchor dump : \n", udf_session->session_num,
					(uint32_t) disc->session_start[udf_session->session_num], (uint32_t) udf_session->session_length);
			udf_dump_descriptor((union dscrptr *) &udf_session->anchor);
			udf_session = udf_session->next;
		};
	);
}


char *udf_messy_unicode_conv(char *buf) {
	static char out_buf[1024];
	uint16_t *uni_pos, uni_char;
	char *pos;

	pos = out_buf;
	uni_pos = (uint16_t *) buf;

	while ((uni_char = *uni_pos++)) {
		if (uni_char & 0xff00) uni_char='_';
		*pos++ = uni_char;
	};

	return out_buf;
}


char *udf_get_osname(int os_class, int os_id) {
	static char buffer[40];

	switch (os_class) {
		case 0 : return "undefined OS";
		case 1 : return "DOS/Windows 3.x";
		case 2 : return "OS/2";
		case 3 : return "MacOS";
		case 4 :
			switch (os_id) {
				case 0 : return "UNIX";
				case 1 : return "IBM AIX";
				case 2 : return "SunOS/Solaris";
				case 3 : return "HP/UX";
				case 4 : return "Silicon Graphics Irix";
				case 5 : return "Linux";
				case 6 : return "MKLinux";
				case 7 : return "FreeBSD/NetBSD?";
				default :
					 sprintf(buffer, "unknown UNIX (%d)", os_id);
					 return buffer;
			};
		case 5 : return "MS Windows 9x";
		case 6 : return "MS Windows NT";
		case 7 : return "OS/400";
		case 8 : return "BeOS";
		case 9 : return "MS Windows CE";
		default :
			break;
	};
	sprintf(buffer, "unknown OS (%d, %d)", os_class, os_id);
	return buffer;
}


void udf_dump_regid(char *prefix, struct regid *id, int regid_type) {
	char    buffer[UDF_REGID_ID_SIZE+1];
	int     cnt, version;
	int8_t *pos;

	memcpy(buffer, id->id, UDF_REGID_ID_SIZE);
	buffer[UDF_REGID_ID_SIZE] = 0;

	printf("%s `%s`", prefix, buffer);
	if (regid_type == UDF_REGID_NAME) {
		printf("\n");
		return;
	};
	printf(" (");
	pos = id->id_suffix;
	switch (regid_type) {
		case UDF_REGID_DOMAIN :
			version = udf_rw16(*(uint16_t *) pos);
			printf("UDFv %x; ", version);
			if ((pos[2]) & UDF_DOMAIN_FLAG_HARD_WRITE_PROTECT) printf("HARD ");
			if ((pos[2]) & UDF_DOMAIN_FLAG_SOFT_WRITE_PROTECT) printf("SOFT");
			if (((pos[2]) & 3) == 0) printf("no");
			printf(" write protect ");
			if ((pos[2]) & ~3) printf("; also undefined flags 0x%d", pos[2] & ~3);
			break;
		case UDF_REGID_UDF :
			version = udf_rw16(*(uint16_t *) pos);
			printf("UDFv %x; ", version);
			printf("%s ", udf_get_osname(pos[2], pos[3]));
			break;
		case UDF_REGID_IMPLEMENTATION :
			printf("%s [", udf_get_osname(pos[0], pos[1]));
			for(cnt=2; cnt < 8; cnt++) {
				printf("%02x ", *pos++);
			};
			printf("]");
			break;
		case UDF_REGID_NAME :
			break;
		case UDF_REGID_APPLICATION :
		default :
			printf("[");
			for(cnt=0; cnt < 8; cnt++) {
				printf("%02x ", *pos++);
			};
			printf("]");
			break;
	};
	printf(") (flags=%d)\n", id->flags);
}


void udf_dump_timestamp(char *prefix, struct timestamp *t) {
	printf("%s (%4d %02d %02d at %02d:%02d:%02d.%02d.%d.%d)\n", prefix, udf_rw16(t->year), t->month, t->day,
			t->hour, t->minute, t->second, t->centisec, t->hund_usec, t->usec);
}


#if 0
void udf_dump_charspec(char *prefix, struct charspec *chsp) {
	int cnt, ch;

	printf("%s type CS%d (", prefix, chsp->type);
	for (cnt=0; cnt<63; cnt++) {
		ch = chsp->inf[cnt];
		if (ch < 32 || ch > 126) {
			printf(".");
		} else {
			printf("%c", ch);
		};
	};
	printf(")\n");
}
#endif


void udf_dump_sparing_table(struct udf_sparing_table *spt) {
	struct spare_map_entry *sp_entry;
	uint32_t entry, entries;

	printf("\t\tSparing table descriptor\n");
	udf_dump_regid("\t\t\tSparing table Id ", &spt->id, UDF_REGID_UDF);
	printf("\t\t\tRelocation table entries          %d\n", udf_rw32(spt->rt_l));
	printf("\t\t\tSequence number                   %d\n", udf_rw32(spt->seq_num));
	printf("\t\t\tMappings :");

	entries = udf_rw32(spt->rt_l);
	for(entry = 0; entry < entries; entry++) {
		if (entry % 4 == 0) printf("\n\t\t\t\t");
		sp_entry = &spt->entries[entry];
		printf("[%08x -> %08x]   ", udf_rw32(sp_entry->org), udf_rw32(sp_entry->map));
	};
	printf("\n");
}


void udf_dump_pri_vol(struct pri_vol_desc *pvd) {
	struct extent_ad *ext;

	printf("\t\tPrimary volume descriptor\n");
	printf("\t\t\tVolume descriptor sequence number %d\n", udf_rw32(pvd->seq_num));
	printf("\t\t\tPrimary volume descriptor number  %d\n", udf_rw32(pvd->pvd_num));
	udf_dump_id("\t\t\tVolume Id     ", 32, pvd->vol_id, &pvd->desc_charset);
	printf("\t\t\tVolume sequence number            %d\n", udf_rw16(pvd->vds_num));
	printf("\t\t\tMaximum volume sequence number    %d\n", udf_rw16(pvd->max_vol_seq));
	printf("\t\t\tInterchange level                 %d\n", udf_rw16(pvd->ichg_lvl));
	printf("\t\t\tMaximum interchange level         %d\n", udf_rw16(pvd->max_ichg_lvl));
	udf_dump_id("\t\t\tVolume set Id ", 128, pvd->volset_id, &pvd->desc_charset);
	// udf_dump_charspec("\t\t\tCharspec for this descriptor     ", &pvd->desc_charset);
	// udf_dump_charspec("\t\t\tCharspec for the explaination    ", &pvd->explanatory_charset);
	ext = &pvd->vol_abstract;
	printf("\t\t\tVolume abstract  at %d for %d bytes\n", udf_rw32(ext->loc), udf_rw32(ext->len));
	ext = &pvd->vol_copyright;
	printf("\t\t\tVolume copyright at %d for %d bytes\n", udf_rw32(ext->loc), udf_rw32(ext->len));
	udf_dump_regid("\t\t\tApplication   id", &pvd->app_id, UDF_REGID_APPLICATION);
	udf_dump_timestamp("\t\t\tTimestamp", &pvd->time);
	udf_dump_regid("\t\t\tImplementator id", &pvd->imp_id, UDF_REGID_IMPLEMENTATION);
	printf("\t\t\tPrevious volume descriptor sequence locator at sector %d\n", udf_rw32(pvd->prev_vds_loc));
	printf("\t\t\tFlags %d\n", udf_rw16(pvd->flags));
}


void udf_dump_implementation_volume(struct impvol_desc *ivd) {
	struct charspec *charspec;

	printf("\t\tImplementation use volume descriptor\n");
	printf("\t\t\tVolume descriptor sequence number %d\n", udf_rw32(ivd->seq_num));
	udf_dump_regid("\t\t\tImplementator identifier", &ivd->impl_id, UDF_REGID_UDF);

	/* check on UDF implementation info ... */
	if (strcmp(ivd->impl_id.id, "*UDF LV Info") == 0) {
		charspec = &ivd->_impl_use.lv_info.lbi_charset;
		// udf_dump_charspec("\t\t\tLV info charspec                 ", charspec);
	        udf_dump_id("\t\t\tLogical volume identifier         ", 128, ivd->_impl_use.lv_info.logvolid, charspec);
	        udf_dump_id("\t\t\tLV info 1                         ",  36, ivd->_impl_use.lv_info.lvinfo1,  charspec);
	        udf_dump_id("\t\t\tLV info 2                         ",  36, ivd->_impl_use.lv_info.lvinfo2,  charspec);
	        udf_dump_id("\t\t\tLV info 3                         ",  36, ivd->_impl_use.lv_info.lvinfo3,  charspec);
		udf_dump_regid("\t\t\tImplementation identifier", &ivd->_impl_use.lv_info.imp_id, UDF_REGID_UDF);
	};
}


void udf_dump_part(struct part_desc *pd) {
	struct part_hdr_desc *part_hdr_desc;

	printf("\t\tPartition descriptor\n");
	printf("\t\t\tVolume descriptor sequence number %d\n", udf_rw32(pd->seq_num));
	printf("\t\t\tFlags                             %d\n", udf_rw16(pd->flags));
	printf("\t\t\tPartition number                  %d\n", udf_rw16(pd->part_num));
	udf_dump_regid("\t\t\tContents", &pd->contents, UDF_REGID_APPLICATION);
	printf("\t\t\tAccessType                        %d\n", udf_rw32(pd->access_type));
	printf("\t\t\tPartition starts at sector %d for %d sectors\n", udf_rw32(pd->start_loc), udf_rw32(pd->part_len));
	udf_dump_regid("\t\t\tImplementator id", &pd->imp_id, UDF_REGID_IMPLEMENTATION);

	printf("\t\t\tPartition contents use (file) descriptors:\n");
	if (strncmp(pd->contents.id, "+NSR0", 5) == 0) {
		part_hdr_desc = (struct part_hdr_desc *) pd->contents_use;
		printf("\t\t\t\tUnallocated space table       at logic block %d for %d bytes\n",
				udf_rw32(part_hdr_desc->unalloc_space_table.lb_num),
				udf_rw32(part_hdr_desc->unalloc_space_table.len)
			);
		printf("\t\t\t\tUnallocated space bitmap      at logic block %d for %d bytes\n",
				udf_rw32(part_hdr_desc->unalloc_space_bitmap.lb_num),
				udf_rw32(part_hdr_desc->unalloc_space_bitmap.len)
			);
		printf("\t\t\t\tPartition integrety table     at logic block %d for %d bytes\n",
				udf_rw32(part_hdr_desc->part_integrety_table.lb_num),
				udf_rw32(part_hdr_desc->part_integrety_table.len)
			);
		printf("\t\t\t\tReusable (freed) space table  at logic block %d for %d bytes\n",
				udf_rw32(part_hdr_desc->freed_space_table.lb_num),
				udf_rw32(part_hdr_desc->freed_space_table.len)
			);
		printf("\t\t\t\tReusable (freed) space bitmap at logic block %d for %d bytes\n",
				udf_rw32(part_hdr_desc->freed_space_bitmap.lb_num),
				udf_rw32(part_hdr_desc->freed_space_bitmap.len)
			);
	} else {
		printf("\t\t\t\tWARNING: Unknown or unused contents\n");
	};
}


void udf_dump_log_vol(struct logvol_desc *lvd) {
	union udf_pmap *pmap;
	uint8_t pmap_type, pmap_size;
	uint8_t *pmap_pos;
	int map, sparing_table;
	uint32_t lb_size, packet_len;

	lb_size = udf_rw32(lvd->lb_size);

	printf("\t\tLogical volume descriptor\n");
	printf("\t\t\tVolume descriptor sequence number %d\n", udf_rw32(lvd->seq_num));
	udf_dump_id("\t\t\tLogical volume id                ",  128, lvd->logvol_id, &lvd->desc_charset);
	printf("\t\t\tLogical block size                %d\n", udf_rw32(lvd->lb_size));
	udf_dump_regid("\t\t\tDomainId", &lvd->domain_id, UDF_REGID_DOMAIN);
	udf_dump_long_ad("\t\t\tFileset descriptor at", &lvd->_lvd_use.fsd_loc);
	printf("\t\t\tMap table length                  %d\n", udf_rw32(lvd->mt_l));
	printf("\t\t\tNumber of part maps               %d\n", udf_rw32(lvd->n_pm));
	udf_dump_regid("\t\t\tImplementation id", &lvd->imp_id, UDF_REGID_IMPLEMENTATION);
	printf("\t\t\tIntegrety sequence at %d for %d bytes\n",
			udf_rw32(lvd->integrity_seq_id.loc), udf_rw32(lvd->integrity_seq_id.len));
	printf("\t\t\tPartion maps follow\n");

	pmap_pos = &lvd->maps[0];
	for(map = 0; map < udf_rw32(lvd->n_pm); map++) {
		pmap = (union udf_pmap *) pmap_pos;
		pmap_type = pmap->data[0];
		pmap_size = pmap->data[1];
		printf("\t\t\t\tPartion map type %d length %d \n", pmap_type, pmap_size);
		/* only pmap types 1 and pmap types 2 are to be used */
		switch (pmap_type) {
			case 1 :
				printf("\t\t\t\t\tVAT and part table at volume sequence number %d partion number %d\n",
						udf_rw16(pmap->pm1.vol_seq_num), udf_rw16(pmap->pm1.part_num));
				break;
			case 2 :
				printf("\t\t\t\t\tVAT and part table at volume sequence number %d partion number %d\n",
						udf_rw16(pmap->pm2.vol_seq_num), udf_rw16(pmap->pm2.part_num));
				udf_dump_regid("\t\t\t\t\tWith id ", &pmap->pm2.part_id, UDF_REGID_UDF);  /* XXX UDF? XXX */
				if (strncmp(pmap->pm2.part_id.id, "*UDF Virtual Partition", UDF_REGID_ID_SIZE) == 0) {
					/* ? */
				};
				if (strncmp(pmap->pm2.part_id.id, "*UDF Sparable Partition", UDF_REGID_ID_SIZE) == 0) {
					packet_len = udf_rw16(pmap->pms.packet_len);
					printf("\t\t\t\t\t\tPacket length                %d sectors (%d bytes)\n",
							packet_len, packet_len * lb_size);
					printf("\t\t\t\t\t\tNumber of sparing tables     %d\n", pmap->pms.n_st);
					printf("\t\t\t\t\t\tSize of each sparing table   %d\n", udf_rw32(pmap->pms.st_size));
					if (pmap->pms.n_st) {
						printf("\t\t\t\t\t\tSparing tables at sectors    ");
						for (sparing_table = 0; sparing_table < pmap->pms.n_st; sparing_table++) {
							printf("%d ", pmap->pms.st_loc[sparing_table]);
						};
						printf("\n");
					};
				};
				break;
			default :
				break;
		};
		pmap_pos += pmap_size;
	};
}


void udf_dump_unalloc_space(struct unalloc_sp_desc *usd) {
	struct extent_ad *alloc_desc;
	int desc_num;

	printf("\t\tUnallocated space descriptor\n");
	printf("\t\t\tVolume descriptor sequence number %d\n", udf_rw32(usd->seq_num));
	printf("\t\t\tNumber of free space slots        %d\n", udf_rw32(usd->alloc_desc_num));
	if (udf_rw32(usd->alloc_desc_num)) {
		printf("\t\t\tFree space at : ");
		for (desc_num = 0; desc_num < udf_rw32(usd->alloc_desc_num); desc_num++) {
			alloc_desc = &usd->alloc_desc[desc_num];
			printf("[%d %d] ", udf_rw32(alloc_desc->loc), udf_rw32(alloc_desc->loc)+udf_rw32(alloc_desc->len));
		};
		printf("\n");
	};
}


void udf_dump_terminating_desc(union dscrptr *desc) {
	desc = desc;

	printf("\t\tTerminating descriptor\n");
}


void udf_dump_logvol_integrity(struct logvol_int_desc *lvid) {
	struct udf_logvol_info *impl;
	uint32_t  part, num_part;
	uint32_t *pos1, *pos2;
	uint32_t  free, size, rest_bytes;
	uint32_t  version;

	printf("\t\tLogical volume integrity descriptor\n");
	udf_dump_timestamp("\t\t\tTimestamp                           ", &lvid->time);
	printf("\t\t\tIntegrity type                       %s\n", udf_rw32(lvid->integrity_type) ? "closed":"open");
	printf("\t\t\tNext integrety sequence at %d for %d bytes\n",
			udf_rw32(lvid->next_extent.loc), udf_rw32(lvid->next_extent.len));

	/* XXX only printing out lower 32 bits ... generic way for printing 64 bits? XXX */
	printf("\t\t\tNext free unique file ID             %d\n", (int32_t) udf_rw64(lvid->logvol_content_use.logvolhdr.next_unique_id));

	printf("\t\t\tLength of implementation use area    %d bytes\n", udf_rw32(lvid->l_iu));

	num_part = udf_rw32(lvid->num_part);
	printf("\t\t\tNumber of partitions                 %d\n", num_part);
	for (part=0; part < num_part; part++) {
		pos1 = &lvid->tables[0] + part;
		pos2 = &lvid->tables[0] + num_part + part;
		free = udf_rw32(*pos1);
		size = udf_rw32(*pos2);
		printf("\t\t\tPartition %d : %d blocks free space out of %d blocks\n", part, free, size);
	};

	/* printout the implementation use field */
	impl = (struct udf_logvol_info *) (lvid->tables + 2*num_part);

	udf_dump_regid("\t\t\tImplemenator Id", &impl->impl_id, UDF_REGID_IMPLEMENTATION );
	printf("\t\t\tNumber of files                      %d\n", udf_rw32(impl->num_files));
	printf("\t\t\tNumber of directories                %d\n", udf_rw32(impl->num_directories));
	version = udf_rw16(impl->min_udf_readver);
	printf("\t\t\tMinimum readversion                  UDFv %x\n", version);
	version = udf_rw16(impl->min_udf_writever);
	printf("\t\t\tMinimum writeversion                 UDFv %x\n", version);
	version = udf_rw16(impl->max_udf_writever);
	printf("\t\t\tMaximum writeversion                 UDFv %x\n", version);
	rest_bytes = udf_rw32(lvid->l_iu)-sizeof(struct udf_logvol_info);
	if (rest_bytes > 0) printf("\t\t\t<%d bytes of undumped extra implementation use area>", rest_bytes);
	printf("\n");
}


void udf_dump_fileset_desc(struct fileset_desc *fsd) {
	printf("\t\tFileset descriptor\n");
	udf_dump_timestamp("\t\t\tTimestamp                         ", &fsd->time);
	printf("\t\t\tInterchange level                  %d\n", udf_rw16(fsd->ichg_lvl));
	printf("\t\t\tMax interchange level              %d\n", udf_rw16(fsd->max_ichg_lvl));
	printf("\t\t\tCharset lists                      %d\n", udf_rw32(fsd->charset_list));
	printf("\t\t\tMax charset lists                  %d\n", udf_rw32(fsd->max_charset_list));
	printf("\t\t\tFileset number                     %d\n", udf_rw32(fsd->fileset_num));
	printf("\t\t\tFileset descriptor number          %d\n", udf_rw32(fsd->fileset_desc_num));
	// udf_dump_charspec("\t\t\tLogical volume id charspec        ", &fsd->logvol_id_charset);
	// udf_dump_charspec("\t\t\tFileset id charspec               ", &fsd->fileset_charset);
	udf_dump_id("\t\t\tLogical volume id                 ", 128, fsd->logvol_id,  &fsd->logvol_id_charset);
	udf_dump_id("\t\t\tFileset id                        ",  32, fsd->fileset_id, &fsd->fileset_charset);
	udf_dump_id("\t\t\tCopyright file id                 ",  32, fsd->copyright_file_id, &fsd->fileset_charset);
	udf_dump_id("\t\t\tAbstract file id                  ",  32, fsd->abstract_file_id,  &fsd->fileset_charset);
	udf_dump_regid("\t\t\tDomainId", &fsd->domain_id, UDF_REGID_DOMAIN);
	udf_dump_long_ad("\t\t\tRootdir ICB found       ", &fsd->rootdir_icb);
	udf_dump_long_ad("\t\t\tNext extend for fileset ", &fsd->next_ex);
	udf_dump_long_ad("\t\t\tStreamdir ICB found     ", &fsd->streamdir_icb); 
}


void udf_dump_fileid_in_charspec(struct fileid_desc *fid, struct charspec *chsp) {
	uint8_t *pos, file_char;

	printf("\tFile id entry\n");
	printf("\t\tFile version number                  %d\n", udf_rw16(fid->file_version_num));
	file_char = fid->file_char;
	printf("\t\tFile characteristics %d :\t", file_char);
	if (file_char & UDF_FILE_CHAR_VIS)  printf("hidden ");
	if (file_char & UDF_FILE_CHAR_DEL)  printf("deleted ");
	if (file_char & UDF_FILE_CHAR_PAR)  printf("parent(..) ");
	if (file_char & UDF_FILE_CHAR_DIR)  printf("directory ");
	if (file_char & UDF_FILE_CHAR_META) printf("METADATA ");
	printf("\n");
	udf_dump_long_ad("\t\tFile ICB", &fid->icb);
	printf("\t\tLength of file identifier area       %d\n", fid->l_fi);
	printf("\t\tOSTA UDF Unique ID                   %d\n", fid->icb.impl.im_used.unique_id);
	printf("\t\tOSTA UDF fileflags                   %d\n", fid->icb.impl.im_used.flags);
	printf("\t\tImplementation use length            %d\n", udf_rw16(fid->l_iu));
	pos = fid->data + udf_rw16(fid->l_iu);
	if (file_char & UDF_FILE_CHAR_PAR) {
		printf("\t\tParent directory ..\n");
	} else {
		udf_dump_id("\t\tFilename", fid->l_fi, pos, chsp);
	};
}


void udf_dump_fileid(struct fileid_desc *fid) {
	struct charspec chsp;

	/* prolly OSTA compressed unicode anyway */
	chsp.type = 0;
	strcpy(chsp.inf, "OSTA Compressed Unicode");

	udf_dump_fileid_in_charspec(fid, &chsp);
}


void udf_dump_icb_tag(struct icb_tag *icb_tag) {
	uint32_t flags, strat_param16;

	flags = udf_rw16(icb_tag->flags);
	strat_param16 = udf_rw16(* (uint16_t *) icb_tag->strat_param);
	printf("\t\tICB Prior direct entries recorded (excl.)   %d\n", udf_rw32(icb_tag->prev_num_dirs));
	printf("\t\tICB Strategy type                           %d\n", udf_rw16(icb_tag->strat_type));
	printf("\t\tICB Strategy type flags                     %d %d\n", icb_tag->strat_param[0], icb_tag->strat_param[1]);
	printf("\t\tICB Number of entries                       %d\n", udf_rw16(icb_tag->max_num_entries));
	printf("\t\tICB     indirect entries/depth              %d\n", strat_param16);
	printf("\t\tICB File type                               %d\n", icb_tag->file_type);
	printf("\t\tICB Parent ICB in logical block %d of mapped partition %d\n",
		udf_rw32(icb_tag->parent_icb.lb_num), udf_rw16(icb_tag->parent_icb.part_num));
	printf("\t\tICB Flags                                   %d\n", udf_rw16(icb_tag->flags));
	printf("\t\t\tFile/directory information using : ");
	switch (flags & UDF_ICB_TAG_FLAGS_ALLOC_MASK) {
		case UDF_ICB_SHORT_ALLOC :
			printf("short allocation descriptor\n");
			break;
		case UDF_ICB_LONG_ALLOC :
			printf("long allocation descriptor\n");
			break;
		case UDF_ICB_EXT_ALLOC :
			printf("extended allocation descriptor (out of specs)\n");
			break;
		case UDF_ICB_INTERN_ALLOC :
			printf("internal in the ICB\n");
			break;
	};
	if (icb_tag->file_type == UDF_ICB_FILETYPE_DIRECTORY)
		if (flags & UDF_ICB_TAG_FLAGS_DIRORDERED)
			printf("\t\t\tOrdered directory\n");
	if (flags & UDF_ICB_TAG_FLAGS_NONRELOC) printf("\t\t\tNot relocatable\n");
	printf("\t\t\tFile flags :");
		if (flags & UDF_ICB_TAG_FLAGS_SETUID) printf("setuid() ");
		if (flags & UDF_ICB_TAG_FLAGS_SETGID) printf("setgid() ");
		if (flags & UDF_ICB_TAG_FLAGS_STICKY) printf("sticky ");
	printf("\n");
	if (flags & UDF_ICB_TAG_FLAGS_CONTIGUES)
		printf("\t\t\tFile is contigues i.e. in one piece effectively \n");
	if (flags & UDF_ICB_TAG_FLAGS_MULTIPLEVERS)
		printf("\t\t\tExpect multiple versions of a file in this directory\n");
}


void udf_dump_indirect_entry(struct indirect_entry *inde) {
	printf("\tIndirect (ICB) entry\n");
	udf_dump_icb_tag(&inde->icbtag);
	udf_dump_long_ad("\t\tPointing at", &inde->indirect_icb);
	printf("\n");
}


void udf_dump_file_entry(struct file_entry *fe) {
	union icb	*icb;
	uint8_t		*pos;
	uint32_t	 length, size, piece_length, piece_flags;
	uint8_t		 addr_type;
	uint32_t	 entry, entries, direct_entries, indirect_entries;
	uint16_t	 strategy, strat_param16;

	//direct_entries = udf_rw32(fe->icbtag.prev_num_dirs);
	strat_param16  = udf_rw16(* (uint16_t *) (fe->icbtag.strat_param));
	entries        = udf_rw16(fe->icbtag.max_num_entries);
	strategy       = udf_rw16(fe->icbtag.strat_type);

	printf("\tFile entry\n");
	udf_dump_icb_tag(&fe->icbtag);
	printf("\t\tUid                                         %d\n", udf_rw32(fe->uid));
	printf("\t\tGid                                         %d\n", udf_rw32(fe->gid));
	printf("\t\tPermissions                                 %x\n", udf_rw32(fe->perm));
	printf("\t\tLink count                                  %d\n", udf_rw16(fe->link_cnt));
	printf("\t\tRecord format                               %d\n", fe->rec_format);
	printf("\t\tRecord display attributes                   %d\n", fe->rec_disp_attr);
	printf("\t\tRecord length                               %d\n", fe->rec_len);
	printf("\t\tInformation length                          %d\n", (uint32_t) udf_rw64(fe->inf_len));
	printf("\t\tLogical blocks recorded                     %d\n", (uint32_t) udf_rw64(fe->logblks_rec));
	udf_dump_timestamp("\t\tAccess time                                ", &fe->atime);
	udf_dump_timestamp("\t\tModification time                          ", &fe->mtime);
	udf_dump_timestamp("\t\tAttribute time                             ", &fe->attrtime);
	printf("\t\tCheckpoint                                         %d\n", udf_rw32(fe->ckpoint));
	udf_dump_long_ad("\t\tExtended attributes ICB at", &fe->ex_attr_icb);
	udf_dump_regid("\t\tImplementation", &fe->imp_id, UDF_REGID_IMPLEMENTATION);
	printf("\t\tUniqueID                                    %d\n", (uint32_t) udf_rw64(fe->unique_id));
	printf("\t\tLength of extended attribute area           %d\n", udf_rw32(fe->l_ea));
	printf("\t\tLength of allocation descriptors            %d\n", udf_rw32(fe->l_ad));
	printf("\t\t<Undumped extended attribute area>\n");
	printf("\t\tAllocation descriptors : \n");
	
	addr_type = udf_rw16(fe->icbtag.flags) & UDF_ICB_TAG_FLAGS_ALLOC_MASK;

	direct_entries = indirect_entries = 0;
	switch (strategy) {
		case	1:
		case	2:
		case	3:
			break;
		case 	4:
			/* just only direct entry level ICBs; no indirect entries */
			/* XXX confusion : only direct entries are allowed but  why is `entries' == 1 ? independent of splitting? XXX */
			if (entries>1) printf("UDF strategy 4 : Entries > 1\n");
			direct_entries   = udf_rw32(fe->l_ad); // entries;
			indirect_entries = 0;
			break;
		case 4096:
			/* one linked list of entries : one direct the other indirect */
			if (strat_param16 != 1)
			direct_entries   = strat_param16;
			indirect_entries = 1;
			break;
		default :
			break;
	};

	entry = 0;
	pos = &fe->data[0] + udf_rw32(fe->l_ea);
	length = udf_rw32(fe->l_ad); size = 0;
	while (length) {
		if (entry % 3 == 0) printf("\n\t");
		printf(" [ ");
		if (entry>direct_entries) printf("indirect ICB "); else printf("direct ");
		/* what to do with strat type == 3 ? or is all set up ok then ? */
		icb = (union icb *) pos;
		switch (addr_type) {
			case UDF_ICB_SHORT_ALLOC  :
				piece_length = udf_rw32(icb->s_ad.len) & (((uint32_t) 1<<31)-1);
				piece_flags  = udf_rw32(icb->s_ad.len) >> 30;		/* XXX ecma167 48.14.1.1 XXX */
				printf("sec %8u for %8d bytes", udf_rw32(icb->s_ad.lb_num), piece_length);
				if (piece_flags) printf(" flags %d", piece_flags);
				size = sizeof(struct short_ad);
				break;
			case UDF_ICB_LONG_ALLOC   :
				piece_length = udf_rw32(icb->l_ad.len) & (((uint32_t) 1<<31)-1);
				piece_flags  = udf_rw32(icb->l_ad.len) >> 30;		/* XXX ecma167 48.14.1.1 XXX */
				printf("sec %8d for %8d bytes, part %d", udf_rw32(icb->l_ad.loc.lb_num), piece_length, 
						udf_rw16(icb->l_ad.loc.part_num));
				if (piece_flags) printf(" flags %d", piece_flags);
				size = sizeof(struct long_ad);
				break;
			case UDF_ICB_EXT_ALLOC    :
				printf("extended alloc (help)");
				size = sizeof(struct ext_ad);
				break;
			case UDF_ICB_INTERN_ALLOC :
				printf("internal here for %d bytes", length);
				size = length;
				break;
		};
		printf(" ] ");
		entry++;
		pos += size;
		length -=size;
	};
	printf("\n");
}


/* dump a space table(entry) descriptor */
void udf_dump_space_entry(struct space_entry_desc *sed) {
	union icb *icb;
	uint32_t addr_type, size, bytes;
	uint32_t piece_sector, piece_length, piece_part;
	uint8_t *pos;

	printf("\tSpace entry table\n");
	udf_dump_icb_tag(&sed->icbtag);
	printf("\t\tSize in bytes                               %d\n", udf_rw32(sed->l_ad));

	pos   = &sed->entry[0];
	bytes = udf_rw32(sed->l_ad);

	addr_type = udf_rw16(sed->icbtag.flags) & UDF_ICB_TAG_FLAGS_ALLOC_MASK;
	while (bytes) {
		size = piece_length = piece_sector = piece_part = 0;
		icb = (union icb *) pos;
		switch (addr_type) {
			case UDF_ICB_SHORT_ALLOC :
				piece_length = udf_rw32(icb->s_ad.len) & (((uint32_t) 1<<31)-1);
				piece_sector = udf_rw32(icb->s_ad.lb_num);
				printf("[at sec %u for %d bytes] ", piece_sector, piece_length);
				size = sizeof(struct short_ad);
				break;
			case UDF_ICB_LONG_ALLOC :
				piece_length = udf_rw32(icb->l_ad.len) & (((uint32_t) 1<<31)-1);
				piece_sector = udf_rw32(icb->l_ad.loc.lb_num);
				piece_part   = udf_rw16(icb->l_ad.loc.part_num);
				size = sizeof(struct long_ad);
				printf("[at sec %u for %d bytes at partition %d] ", piece_sector, piece_length, piece_part);
				break;
			case UDF_ICB_EXT_ALLOC :
			case UDF_ICB_INTERN_ALLOC :
				printf("\t\t\tWARNING : an internal alloc in a space entry?\n");
				return;
		};
		bytes -= size;
	};
}


/* dump a space bitmap descriptor */
void udf_dump_space_bitmap(struct space_bitmap_desc *sbd) {
	uint32_t bits, from, now, cnt;
	uint8_t byte, bit, bitpos, state, *pos;

	printf("\tSpace bitmap\n");
	printf("\t\tNumber of bits                              %d\n", udf_rw32(sbd->num_bits));
	printf("\t\tNumber of bytes                             %d\n", udf_rw32(sbd->num_bytes));
	printf("\t\tMarked parts at :\n");

	pos = sbd->data;
	bits = udf_rw32(sbd->num_bits);

	/* shield */
	if (bits > 2000*8) bits = 2000*8;

	printf("\t\t\t");
	cnt = 0; from = 0; now = 0; bitpos = 0; byte = *pos; state = byte & 1;
	while (now < bits) {
		if (bitpos == 0) {
			byte = *pos++;
		};
		bit = byte & 1;
		if (bit != state) {
			if (state) {
				printf("[%d-%d] ", from, now);
				if (cnt % 8 == 7) printf("\n\t\t\t");
				cnt++;
			};
			from = now+1;
			state = bit;
		};
		byte >>= 1;
		bitpos = (bitpos+1) & 7;
		now++;
	};
	if (state) printf("[%d-%d]", from, now);
	if (bits < udf_rw32(sbd->num_bits)) printf(" .... <trimmed>\n");
}


/* main descriptor `dump' function */
void udf_dump_descriptor(union dscrptr *dscrpt) {
	struct desc_tag *tag = &dscrpt->tag;

	/* check if its a valid descritor */
	if (udf_rw16(tag->id == 0) && udf_rw16(tag->descriptor_ver) == 0) return;

	udf_dump_desc(tag);

	errno = udf_check_tag(dscrpt);
	if (errno) {
		printf("\tBAD TAG\n");
		return;
	};
	switch (udf_rw16(tag->id)) {
		case TAGID_SPARING_TABLE :
			udf_dump_sparing_table(&dscrpt->spt);
			break;
		case TAGID_PRI_VOL :
			udf_dump_pri_vol(&dscrpt->pvd);
			break;
		case TAGID_ANCHOR :
			udf_dump_anchor(&dscrpt->avdp);
			break;
		case TAGID_VOL :
			udf_dump_unimpl(dscrpt);
			break;
		case TAGID_IMP_VOL :
			udf_dump_implementation_volume(&dscrpt->ivd);
			break;
		case TAGID_PARTITION :
			udf_dump_part(&dscrpt->pd);
			break;
		case TAGID_LOGVOL :
			udf_dump_log_vol(&dscrpt->lvd);
			break;
		case TAGID_UNALLOC_SPACE :
			udf_dump_unalloc_space(&dscrpt->usd);
			break;
		case TAGID_TERM :
			udf_dump_terminating_desc(dscrpt);
			break;
		case TAGID_LOGVOL_INTEGRITY :
			udf_dump_logvol_integrity(&dscrpt->lvid);
			break;
		case TAGID_FSD :
			udf_dump_fileset_desc(&dscrpt->fsd);
			break;
		case TAGID_FID :
			udf_dump_fileid(&dscrpt->fid);
			break;
		case TAGID_ALLOCEXTENT :
			udf_dump_unimpl(dscrpt);
			break;
		case TAGID_INDIRECT_ENTRY :
			udf_dump_indirect_entry(&dscrpt->inde);
			break;
		case TAGID_FENTRY :
			udf_dump_file_entry(&dscrpt->fe);
			break;
		case TAGID_EXTATTR_HDR :
			udf_dump_unimpl(dscrpt);
			break;
		case TAGID_UNALL_SP_ENTRY :
			udf_dump_space_entry(&dscrpt->sed);
			break;
		case TAGID_SPACE_BITMAP :
			udf_dump_space_bitmap(&dscrpt->sbd);
			break;
		case TAGID_PART_INTEGRETY :
			udf_dump_unimpl(dscrpt);
			break;
		default :
			break;
	};
	printf("\n");
}


/* this one is special since the VAT table has no tag but is a file */
void udf_dump_vat_table(struct udf_part_mapping *udf_part_mapping) {
	struct charspec  chsp;
	struct udf_vat  *vat;
	uint32_t	 previous_vat, entry, vat_entries, *vat_pos, version;
	uint8_t		*pos;

	/* prolly OSTA compressed unicode anyway */
	chsp.type = 0;
	strcpy(chsp.inf, "OSTA Compressed Unicode");

	vat = udf_part_mapping->vat;
	printf("\tVAT table: ");
	printf("%s UDF 2.00 format\n", vat?"post":"pre");

	if (vat) {
		printf("\t\tHeader length                        %d\n", udf_rw16(vat->header_len));
		printf("\t\tImplementation use length            %d\n", udf_rw16(vat->impl_use_len));
		udf_dump_id("\t\tLogical volume id                   ", 128, vat->logvol_id,  &chsp);
		printf("\t\tNumber of files                      %d\n", udf_rw32(vat->num_files));
		printf("\t\tNumber of directories                %d\n", udf_rw32(vat->num_directories));
		version = udf_rw16(vat->min_udf_readver);
		printf("\t\tMinimum readversion                  UDFv %x\n", version);
		version = udf_rw16(vat->min_udf_writever);
		printf("\t\tMinimum writeversion                 UDFv %x\n", version);
		version = udf_rw16(vat->max_udf_writever);
		printf("\t\tMaximum writeversion                 UDFv %x\n", version);
		if (udf_rw16(vat->impl_use_len)) printf("\t\t<undumped implementation use area>");
		pos = &vat->data[0] + udf_rw16(vat->impl_use_len);
		vat_pos = (uint32_t *) pos;
		vat_entries = (udf_rw16(vat->header_len) - udf_rw16(vat->impl_use_len))/4;	/* definiton */
		previous_vat = udf_rw32(vat->prev_vat);
	} else {
		vat_entries = udf_part_mapping->vat_entries;
		pos = udf_part_mapping->vat_translation;
		vat_pos = (uint32_t *) pos;
		udf_dump_regid("\t\tIdentifier id (can be wrong)        ", (struct regid *) (vat_pos+vat_entries), UDF_REGID_NAME);
		previous_vat = udf_rw32(*(vat_pos + vat_entries + 32/4));			/* definition */
	};
	if (previous_vat == 0xffffffff) {
		printf("\t\tNo previous VAT recorded\n");
	} else {
		printf("\t\tPrevious VAT recorded at offset      %d\n", previous_vat);
	};

	printf("\t\tNumber of VAT entries                %d\n", vat_entries);
	printf("\t\tVAT dump :");
	for (entry=0; entry < vat_entries; entry++) {
		if ((entry % 4) == 0) printf("\n\t");
		printf("[0x%08x -> 0x%08x] ", entry, *vat_pos++);
	};
	printf("\n");
}


void udf_dump_volumeset_info(struct udf_volumeset *udf_volumeset) {
	struct udf_pri_vol	*primary;
	struct udf_log_vol	*logical;
	struct udf_partition	*udf_partition;
	struct udf_discinfo	*disc;
	char			*name;
	int			 num_volumes, num_partitions;
	int			 subvolume, part_num, track_num;

	num_volumes = 0;	/* shut up gcc */
	if (udf_volumeset->obsolete) return;

	primary = udf_volumeset->primaries;
	if (primary) {
		num_volumes =  udf_rw16(primary->pri_vol->max_vol_seq);
		if (udf_volumeset->obsolete) printf("OBSOLETE\n");	/* XXX */

		printf("Volume set ");
		udf_dump_id(NULL, 32, primary->pri_vol->volset_id, &primary->pri_vol->desc_charset);
		printf(" (%d volume%s) ", num_volumes, num_volumes>1?"s":"");

		num_partitions = udf_volumeset->max_partnum;
		printf("with %d partition%s\n", num_partitions, num_partitions>1?"s":"");

		/* better loop trough the partition numbers to display them in a defined order */
		for(part_num = 0; part_num < num_partitions; part_num++) {
			udf_partition = udf_volumeset->parts;
			while (udf_partition && (udf_rw16(udf_partition->partition->part_num) != part_num))
				udf_partition = udf_partition->next;
			if (udf_partition) {
				/* there is information */
				assert(udf_partition->udf_session);
				assert(udf_partition->udf_session->disc);
				assert(udf_partition->partition);
				assert(part_num == udf_rw16(udf_partition->partition->part_num));
				track_num = udf_partition->udf_session->session_num;
				disc      = udf_partition->udf_session->disc;

				printf("\tPartition number %d at device `%s' session %d from sector %d(+%d) for %d sectors\n",
						part_num,
						disc->dev_name,
						track_num,
						udf_rw32(udf_partition->partition->start_loc),
						udf_partition->udf_session->session_offset,
						udf_rw32(udf_partition->partition->part_len)
				      );
			} else {
				printf("\tUnknown partition %d [unknown]\n", part_num);
			};
		};
	};
	while (primary) {
		subvolume = udf_rw16(primary->pri_vol->vds_num);

		printf("\tPrimary volume ");
		udf_dump_id(NULL, 32, primary->pri_vol->vol_id, &primary->pri_vol->desc_charset);
		printf(" (part %d/%d) ", subvolume, num_volumes);

		printf("created by `%s' ", primary->pri_vol->imp_id.id);
		if (*primary->pri_vol->app_id.id)
			printf("for application `%s' ",primary->pri_vol->app_id.id);
		printf("\n");

		logical = primary->log_vols;
		if (logical) {
			while (logical) {
				name = logical->log_vol->logvol_id;
				udf_dump_id("\t\tcontains logical volume ", 128, name, &logical->log_vol->desc_charset);
				if (logical->broken) printf("\t\t\tBROKEN\n");

				logical = logical->next;
			};
		};
		printf("\n");
		primary = primary->next;
	};
}


void udf_dump_alive_sets(void) {
	struct udf_volumeset *udf_volumeset;

	printf("UDF volume sets marked alive :\n");
	udf_volumeset = udf_volumeset_list;
	while (udf_volumeset) {
		udf_dump_volumeset_info(udf_volumeset);
		udf_volumeset = udf_volumeset->next;
	};
	printf("\n");
}


void udf_dump_file_entry_node(struct udf_icb_node *udf_icb_node, char *name, int recurse) {
	struct udf_readdir_stream *readdir;
	struct udf_icb_node		  *sub_node;
	struct dirent		   entry;
	struct long_ad		   icb;
	char fullpath[1024];	/* XXX arbitrary length XXX */
	int  udf_type, new_recurse;

	if (!udf_icb_node) return;
	UDF_VERBOSE_TABLES(printf("\tUDF file entry dump"));

	if (udf_icb_node->dscrptr->fe.icbtag.file_type == UDF_ICB_FILETYPE_DIRECTORY && recurse) {
		errno = udf_opendir(udf_icb_node, 0, &readdir);
		if (!errno) {
			errno = udf_readdir_entry(readdir, &entry, &icb, &udf_type);
			while (!readdir->eof) {
				if (errno) perror("\treaddir_entry :");
				new_recurse = (strcmp(entry.d_name, ".") && strcmp(entry.d_name, ".."));
				errno = udf_lookup_node(udf_icb_node->udf_log_vol, &icb, &sub_node);
				if (errno) perror("\tlookup_node :");
				sprintf(fullpath, "%s/%s", name, entry.d_name);
				udf_dump_file_entry_node(sub_node, fullpath, new_recurse);

				errno = udf_readdir_entry(readdir, &entry, &icb, &udf_type);
			};
			udf_closedir(readdir);
		} else {
			perror("\topening directory :");
		};
		return;
	};
#if 1
	printf("\t%s\n", name);
#else
{
	uint8_t *file_data;
	uint64_t file_data_got_len, file_data_total_len, data_pos;
	static int cnt=0;
	int fileh, len;
	char buf[1024];

	if (udf_icb_node->dscrptr->fe.icbtag.file_type == UDF_ICB_FILETYPE_DIRECTORY) {
		sprintf(buf, "grab/%s", name);
		len = strlen(buf);
		if (strcmp(buf+len-2, "/.")  == 0) buf[len-2] = 0;
		if (strcmp(buf+len-3, "/..") == 0) return;
		mkdir(buf, 0700);
		if (errno) perror("creating dir");
		return;
	};

	sprintf(buf, "grab/%s", name);

	fileh = open(buf, O_WRONLY | O_CREAT, 0666);
	if (fileh>=0) {
		data_pos = file_data_got_len = file_data_total_len = 0;
		do {
			errno = udf_read_file_part(udf_icb_node, data_pos, 1024*1024, (void **) &file_data, &file_data_got_len, &file_data_total_len);
			data_pos += file_data_got_len;
			write(fileh, file_data, file_data_got_len);
			free(file_data);
if (file_data_total_len) {
printf("\r\t%-70s %10d / %10d bytes (%3d%%) successfully", name, (uint32_t) data_pos, (uint32_t) file_data_total_len, (int) (100.0*(float) data_pos / file_data_total_len));
fflush(stdout);
};
			if (errno) {
				perror("Got this error while retrieving file :");
				printf("\n\n\n\n\n");	/* XXX */
				break;
			};
		} while ((data_pos < file_data_total_len));
		printf("\n");
		close(fileh);
		cnt++;
	} else {
		printf("Help! can't open file %s for output\n", buf);
	};
};
#endif
}


void udf_dump_root_dir(struct udf_mountpoint *mountpoint) {
	struct udf_icb_node *udf_icb_node;

	printf("Root dir dump\n");
	errno = udf_lookup_node(mountpoint->udf_log_vol, &mountpoint->fileset_desc->rootdir_icb, &udf_icb_node);
	if (errno) perror("\t");
	if (mountpoint->fileset_desc->rootdir_icb.len) udf_dump_file_entry_node(udf_icb_node, ":Rootdir", 1);

	printf("Streamdir dump\n");
	errno = udf_lookup_node(mountpoint->udf_log_vol, &mountpoint->fileset_desc->streamdir_icb, &udf_icb_node);
	if (errno) perror("\t");
	if (mountpoint->fileset_desc->streamdir_icb.len) udf_dump_file_entry_node(udf_icb_node, ":Streamdir", 1);
}


