/* $NetBSD$ */

/*
 * File "udfdump.c" is part of the UDFclient toolkit.
 * File $Id: udfdump.c,v 1.46 2003/09/14 14:19:03 imago Exp $ $Name: release-ro1 $
 *
 * Copyright (c) 2003 Reinoud Zandijk <reinoud@netbsd.org>
 * All rights reserved.
 *
 * The UDFclient toolkit is distributed under the Clarified Artistic Licence.
 * A copy of the licence is included in the distribution as
 * `LICENCE.clearified.artistic' and a copy of the licence can also be
 * requested at the GNU foundantion's website.
 *
 * Visit the UDFclient toolkit homepage http://www.13thmonkey.org/udfclient/
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <fcntl.h>
#include <unistd.h>
#include "errno.h"
#include "udf.h"
#include "udf_bswap.h"


/* switches */

//#define DEBUG(a) (a)
#define DEBUG(a) if (0) { a; };
int dump_sequential	 = 0;
int dump_directory_trees = 0;


/* include the dump parts ... in order to get a more sane splitting */
extern void udf_dump_descriptor(union dscrptr *dscrpt);
extern void udf_dump_alive_sets(void);
extern void udf_dump_root_dir(struct udf_mountpoint *mountpoint);


/* globals */
extern int errno;
extern int udf_verbose;
extern int scsilib_verbose;


/* main discdump routine */

void udf_dump_sequential(char *dev_name) {
	struct udf_discinfo *disc;
	union dscrptr		*dscr;
	udf_sector		 sector;
	uint64_t		 sec_num;
	int			 error;

	udf_verbose     = UDF_VERBLEV_ACTIONS;
	scsilib_verbose = 1;

	printf("Opening disc:\n");
	errno = udf_open_disc(dev_name, &disc);
	if (errno) return;

	printf("Sequential dump of device/file\n");
	sec_num = 0;
	do {
		error = udf_read_phys(&sector, disc, sec_num, "sector");
		dscr = (union dscrptr *) sector;
		printf("%8d ", (int) sec_num); udf_dump_descriptor(dscr);
		sec_num++;
	} while (!error);
}


void udf_dump_disc(struct udf_discinfo *disc, int min_session_number, int max_session_number) {
	struct hash_entry     *mountpoint_entry;
	struct udf_mountpoint *mountpoint;

	printf("Dumping directory tree for each non obsolete logical volume's fileset\n");
	mountpoint_entry = udf_mountables->seq_head;
	while (mountpoint_entry) {
		mountpoint = (struct udf_mountpoint *) mountpoint_entry->entry;

		printf("\n");
		printf("Directory tree of mount point %s\n", mountpoint->mount_name);
		udf_dump_root_dir(mountpoint);

		mountpoint_entry = mountpoint_entry->seq_next;
	};
}


int usage(char *program) {
	fprintf(stderr, "Usage %s [options] devicename [devicename]*)\n", program);
	fprintf(stderr, "-S		dump device sequential\n"
			"-t		dump complete directory contents\n"
			"-u level	UDF system verbose level\n"
			"-s		verbose SCSI command errors\n"
			"-r range	only use UDF sessions range like -3,4-5 or 6-\n"
	);
	return 1;
}


int main(int argc, char *argv[]) {
	struct udf_discinfo *disc, *next_disc;
	char flag, *progname, *range;
	int min_session_number, max_session_number;

	progname = argv[0];
	if (argc == 1) return usage(progname);

	dump_sequential      = 0;
	dump_directory_trees = 0;
	range                = NULL;
	while ((flag = getopt(argc, argv, "u:r:sSt")) != -1) {
		switch (flag) {
			case 'u' :
				udf_verbose = atoi(optarg);
				break;
			case 'r' :
				range = strdup(optarg);
				if (udf_check_session_range(range)) {
					fprintf(stderr, "Invalid range %s\n", range);
					return usage(progname);
				};
				break;
			case 's' :
				scsilib_verbose = 1;
				break;
			case 'S' :
				dump_sequential = 1;
				break;
			case 't' :
				dump_directory_trees = 1;
				break;
			default  :
				return usage(progname);
		};
	};
	argv += optind;
	argc -= optind;

	if (dump_sequential) {
		if (argc != 1) return usage(progname);
		udf_dump_sequential(*argv);
		return 0;
	};

	if (argc == 0) return usage(progname);

	min_session_number = 0;
	max_session_number = 999;

	/* all other arguments are devices */
	udf_discs_list = NULL;
	while (argc) {
		printf("Opening device %s\n\n", *argv);
		errno = udf_open_and_read_disc(*argv, range, &disc);
		if (errno) {
			perror("Can't open my device; bailing out ");
			exit(1);
		};

		if (dump_directory_trees) udf_dump_disc(disc, min_session_number, max_session_number);

		disc->next = udf_discs_list;
		udf_discs_list = disc;
		argv++; argc--;
		if (udf_verbose) printf("\n\n");
	};

	printf("Resulting list of alive sets :\n\n");
	udf_dump_alive_sets();

	printf("Closing discs\n");
	disc = udf_discs_list;
	while (disc) {
		next_disc = disc->next;
		udf_close_disc(disc);
		disc = next_disc;
	};

	return 0;
}

